
-- Set 'require' to look in the program root folder first, then in the
-- current folder
package.path = eihort.ProgramPath .. "?.lua;?.lua;;";
MinecraftJar = eihort.MinecraftPath .. "bin/minecraft.jar";

require "mapview"
require "worldmenu"

------------------------------------------------------------------------------
-- Event dispatch system

EventDownAction = { };
EventUpAction = { };
EventState = { };

local function eventDown( evt )
	EventState[evt] = false;
	local action = EventDownAction[evt];
	if action then
		action();
	end
end
local function eventUp( evt )
	EventState[evt] = nil;
	local action = EventUpAction[evt];
	if action then
		action();
	end
end
Event = {
	-- The names of these events are (mostly) hard-coded in eihort.exe

	-- To be filled in later:
	-- redraw(): Called to draw the screen again
	-- mousemove( x, y ): The mouse just moved
	-- idle(): Called when there's nothing else to do.
	--         Return true to save CPU

	active = function( state, ... )
		local types = { ... };
		for _, type in ipairs( types ) do
			if type == 'app' then
				if state then
					Event.resize( eihort.getWindowDims() );
				end
			elseif type == 'mouse' then
				if state then
					eventDown( 'mfocus' );
				else
					eventUp( 'mfocus' );
				end
			end
		end
	end;
	resize = function( w, h )
		ScreenWidth = w;
		ScreenHeight = h;
		ScreenAspect = w / h;
		eventDown( 'resize' );
	end;
	keydown = eventDown;
	keyup = eventUp;
	mousedown = function( button, x, y )
		Event.mousemove( x, y );
		eventDown( 'mouse' .. button );
	end;
	mouseup = function( button, x, y )
		Event.mousemove( x, y );
		eventUp( 'mouse' .. button );
	end;
	quit = function()
		-- User closed the window
		QuitFlag = 0;
	end;
	error = function( msg )
		local ignore = eihort.errorDialogYesNo( "Runtime Error", debug.traceback( msg, 3 ) .. "\n\nIgnore the error?" );
		if not ignore then
			QuitFlag = 1;
		end
	end;
};

local function main()
	-- Helper to process a single event
	local function processEvent( evtStr, ... )
		if not evtStr then
			return false;
		end

		local handler = Event[evtStr];
		if handler then
			handler( ... );
		end

		return true;
	end

	-- The main loop
	assert( Event.redraw );
	assert( Event.idle );

	while not QuitFlag do
		if not processEvent( eihort.pollEvent() ) then
			local idle = Event.idle();
			if eihort.shouldRedraw() then
				Event.redraw();
			end
			if idle then
				eihort.yield(); -- Don't kill the CPU
			end
		end
	end
	return QuitFlag;
end

------------------------------------------------------------------------------
-- Read in the config

Config = { };
do
	local configChunk = assert( loadfile( eihort.ProgramPath .. "eihort.config" ) );
	setfenv( configChunk, Config );
	local success = xpcall( configChunk, function( msg )
		eihort.errorDialog( "Error in eihort.config", msg )
	end );
	if not success then
		return 1;
	end
end

------------------------------------------------------------------------------
-- World loading function

function loadWorld( path )
	local levelDatPath = path .. "/level.dat";
	local f = io.open( levelDatPath, "rb" );
	if f then
		f:close();
		local oName, dat = eihort.loadNBT( levelDatPath );
		if dat then
			local levelData = dat:get("Data");
			if levelData then
				local version = levelData:get("version");
				return eihort.loadWorld( path, version == 0x4abd );
			end
			dat:destroy();
		end
	end
end

------------------------------------------------------------------------------
-- Process the command line

local argv = { ... };
local startWorld;
if #argv > 0 then
	startWorld = loadWorld( argv[1] );
	if startWorld:getRegionCount() == 0 then
		eihort.errorDialog( "Eihort Error", "No regions were found at '" .. argv[1] .. "'.\n\nThere does not appear to be a Minecraft world at that path." );
		return 1;
	end
end

------------------------------------------------------------------------------
-- Initialize video

ScreenWidth = Config.screenwidth or 800;
ScreenHeight = Config.screenheight or 600;
ScreenAspect = ScreenWidth / ScreenHeight;
assert( eihort.initializeVideo( ScreenWidth, ScreenHeight, Config.fullscreen, Config.multisample or 0 ) );

if startWorld then
	-- World loaded - enter the main map view
	beginMapView( startWorld );
else
	-- Show the world selection menu first
	beginWorldMenu();
end

------------------------------------------------------------------------------
-- Error trap

while true do
	local success, ret = xpcall( main, function( ... ) return Event.error( ... ); end );
	if success then
		-- The return value of this function is the return value of eihort.exe
		return ret;
	end
end

