
require "ui"
require "blockids"
require "biomes"

-- Block highlights
local blockHighlights = {
	--{ "name", id1, id2, ... }
	{ "Coal Ore", 16 };
	{ "Iron Ore", 15 };
	{ "Gold Ore", 14 };
	{ "Diamond", 56 };
	{ "Redstone Ore", 73, 74 };
	{ "Lapis Lazuli Ore", 21 };
	{ "Clay", 82 };
	{ "Moss Stone", 48 };
	{ "Rails", 66, 27, 28 };
	{ "Redstone Wire", 55, 75 };
	{ "Trees", 17, 18 };
	{ "Signs", 63, 78 };
	{ "Water", 8, 9 };
	{ "Chests", 54, 95 };
	{ "Spawner", 52 };
	{ "Portal", 90 };
};

-- Helper functions
local function takeScreenshotNow()
	local fnroot = os.date( "./eihort-%Y.%m.%d-%H.%M.%S" );
	local fn = fnroot .. ".png";
	local f = io.open( fn );
	local i = 0;
	while f do
		f:close();
		fn = fnroot .. "-" .. i .. ".png";
		i = i + 1;
		f = io.open( fn );
	end
	local screen = eihort.screengrab();
	local success, msg = screen:writePNG( fn );
	if not success then
		eihort.errorDialog( "Screenshot failed", msg );
	end
end

local function setGpuAllowance( view )
	local allowance = Config.max_gpu_mem or 0;
	if allowance == 0 then
		allowance = eihort.getVideoMem();
		if not allowance or allowance == 0 then
			eihort.errorDialog( "VRAM Autodetect", "Could not detect the amount of video memory available. Please set max_gpu_mem in eihort.config to an appropriate value. Defaulting to 512MB." );
			allowance = 512*1024*1024;
		end
		allowance = allowance * 0.9;
	else
		allowance = allowance * 1024 * 1024;
	end
		
	view:setGpuAllowance( allowance );
end

local function moveSpawnHere( world, dim, x, y, z )
	if dim ~= 0 then
		eihort.errorDialog( "Move spawn", "The spawn must be located in the overworld." );
		return;
	end

	local levelDatPath = world:getRootPath() .. "level.dat";
	local rootName, levelDat = eihort.loadNBT( levelDatPath );
	if not levelDat then
		eihort.errorDialog( "Move spawn", "Failed to open level.dat." );
		return;
	end

	local levelData = levelDat:get("Data");
	if not levelData then
		eihort.errorDialog( "Move spawn", "level.dat doesn't have TAG_Compound \"Data\"?" );
		return;
	end

	levelData.set( "SpawnX", z, "int" );
	levelData.set( "SpawnY", y, "int" );
	levelData.set( "SpawnZ", x, "int" );

	levelDat:write( levelDatPath, rootName );
	levelDat:destroy();
end

local function movePlayerHere( world, dim, x, y, z )
	local levelDatPath = world:getRootPath() .. "level.dat";
	local rootName, levelDat = eihort.loadNBT( levelDatPath );
	if not levelDat then
		eihort.errorDialog( "Move player", "Failed to open level.dat." );
		return;
	end

	local levelData = levelDat:get("Data");
	if not levelData then
		eihort.errorDialog( "Move player", "level.dat doesn't have TAG_Compound \"Data\"?" );
		return;
	end
	local player = levelData:get("Player");
	if not player then
		player = levelData:newCompound( "Player" );
	end

	local pos = player:newList( "Pos", "double" );
	pos[1] = z;
	pos[2] = y;
	pos[3] = x;

	player:set( "Dimension", dim, "int" );

	levelDat:write( levelDatPath, rootName );
	levelDat:destroy();
end


-- Main Map view stuff
function beginMapView( world )
	local worldPath = world:getRootPath();
	local blocks = loadBlockDesc();
	loadBiomeTextures( blocks, world:getRootPath() );
	local worldView = world:createView( blocks, Config.qtree_leaf_size or 7 );
	local sky = eihort.newSky();
	setGpuAllowance( worldView );

	local mouseX, mouseY = eihort.getMousePos();
	local ignoreNextMM = false;
	local mouseLook = false;
	local controlsOn = 0;
	eihort.showCursor( true );

	local worldTime, viewDistance = Config.start_time or 0, Config.start_view_distance or 500;
	local pitch, azimuth;
	local lightBase, lightBaseStr = 0.8, "Overworld";
	local showUI, uiRoot = true, ui.newRoot();
	local infoDisplay = ui.newLabel( "", 0.03, 0.02, 0, 0, 1, 1 );
	local takeScreenshot = false;
	local speedModifier = 1;
	local eyeX, eyeY, eyeZ, fwdX, fwdY, fwdZ, upX, upY, upZ, rightX, rightY, rightZ;
	local fov = (Config.fov or 75) * math.pi / 180;
	local nearPlane = 0.1;
	local monitorEnabled = false;
	
	local execNoLoad = { };
	local function execWhenNothingIsLoading( what, a, b, c, d )
		table.insert( execNoLoad, function() what( a, b, c, d ); end );
		worldView:pauseLoading( true );
	end

	local function refreshPosition()
		fwdX, fwdY, fwdZ, upX, upY, upZ, rightX, rightY, rightZ = eihort.fwdUpRightFromAzPitch( azimuth, pitch );
		worldView:setPosition( eyeX, eyeY, eyeZ, fwdX, fwdY, fwdZ, upX, upY, upZ );
		eihort.shouldRedraw( true );
	end
	local function resetPosition()
		eyeX, eyeY, eyeZ = Config.start_x or 0, Config.start_y or 100, Config.start_z or -20;
		pitch = Config.start_pitch or -1;
		azimuth = Config.start_azimuth or 0;
		refreshPosition();
	end
	local function refreshInfoDisplay()
		local triCount, vtxMem, idxMem, texMem = worldView:getLastFrameStats();
		infoDisplay.text = string.format( "Coords: (%.0f %.0f %.0f)\nTriangles: %u\nMem: %.0f MB (%.0f free)\nSee eihort.config for key bindings.\n\nLight Strength: %s\nTime: %.0f:%02.0f\nView Distance: %.0f%s",
			eyeZ, eyeY, eyeX,
			triCount,
			(vtxMem + idxMem + texMem) / (1024*1024),
			worldView:getGpuAllowanceLeft() / (1024*1024),
			lightBaseStr,
			math.floor( worldTime * 12 + 12 ), math.floor( math.fmod( worldTime * 12 + 12, 1 ) * 60 ),
			viewDistance,
			(monitorEnabled and "\nMonitoring World Folder") or "" );
	end
	local function resetCameraParams()
		worldView:setCameraParams( fov, ScreenWidth / ScreenHeight, nearPlane, viewDistance * 1.05 );
	end

	resetPosition();
	refreshInfoDisplay();
	resetCameraParams();
	worldView:setViewDistance( viewDistance );
	worldView:setOutsideLightLevel( sky:setTime( worldTime ) );
	worldView:setLightBase( lightBase );
	EventDownAction = { };
	EventUpAction = { };

	-------------------------------------------------------------------
	-- Available impulse actions

	local impulseActions = {
		reloadworld = function()
			worldView:reloadAll();
		end;
		togglemonitor = function()
			monitorEnabled = not monitorEnabled;
			world:setMonitorState( monitorEnabled );
			eihort.shouldRedraw( true );
		end;
		changelightmode = function()
			if lightBase < 0.85 then
				lightBase = 0.9;
				lightBaseStr = "Nether";
			elseif lightBase < 0.95 then
				lightBase = 1.0;
				lightBaseStr = "Full Bright";
			else
				lightBase = 0.8;
				lightBaseStr = "Overworld";
			end
			worldView:setLightBase( lightBase );
			eihort.shouldRedraw( true );
		end;
		resetposition = resetPosition;
		toggleui = function()
			showUI = not showUI;
			if showUI then
				uiRoot:mouseEnter( mouseX / ScreenWidth, mouseY / ScreenHeight );
			else
				uiRoot:mouseLeave();
			end
		end;
		screenshot = function()
			takeScreenshot = true;
			eihort.shouldRedraw( true );
		end;
		freemouse = function()
			if mouseLook then
				mouseLook = false;
				eihort.showCursor( true );
				if showUI then
					uiRoot:mouseEnter( mouseX / ScreenWidth, mouseY / ScreenHeight );
				end
			end
		end;
		speedup = { function() speedModifier = Config.speedup_modifier or 8; end;
		            function() speedModifier = 1; end };
		slowdown = { function() speedModifier = Config.slowdown_modifier or 1/8; end;
		             function() speedModifier = 1; end };
	};

	-------------------------------------------------------------------
	-- Key binding setup
	local controls = {
		moveforward = 0;
		moveback = 0;
		moveright = 0;
		moveleft = 0;
		moveup = 0;
		movedown = 0;
		['viewdist-'] = 0;
		['viewdist+'] = 0;
		['time-'] = 0;
		['time+'] = 0;
	};
	for k, v in pairs( Config.keys ) do
		local imp = impulseActions[v];
		if imp then
			if type(imp) == "table" then
				EventDownAction[k] = imp[1];
				EventUpAction[k] = imp[2];
			else
				EventDownAction[k] = imp;
				EventUpAction[k] = nil;
			end
		elseif controls[v] then
			EventDownAction[k] = function()
				controlsOn = controlsOn + 1;
				controls[v] = controls[v] + 1;
			end;
			EventUpAction[k] = function()
				controlsOn = controlsOn - 1;
				controls[v] = controls[v] - 1;
			end;
		else
			error( "Unknown key binding: " .. v );
		end
	end

	-------------------------------------------------------------------
	-- Mouse stuff
	Event.mousemove = function( x, y )
		if ignoreNextMM then
			ignoreNextMM = false;
			return;
		end

		if mouseLook then
			local dx = x - mouseX;
			local dy = y - mouseY;

			azimuth = azimuth - dx * (Config.mouse_sensitivity_x or 0.008);
			pitch = pitch - dy * (Config.mouse_sensitivity_y or 0.01);
			if azimuth > math.pi then
				azimuth = azimuth - 2 * math.pi;
			elseif azimuth < -math.pi then
				azimuth = azimuth + 2 * math.pi;
			end
			if pitch > math.pi / 2 then
				pitch = math.pi / 2;
			elseif pitch < -math.pi / 2 then
				pitch = -math.pi / 2;
			end

			refreshPosition();
			ignoreNextMM = true;
			eihort.warpMouse( mouseX, mouseY );
			return;
		end

		mouseX = x;
		mouseY = y;

		-- Update the UI
		if showUI then
			uiRoot:mouseMove( x / ScreenWidth, y / ScreenHeight );
		end
	end;
	EventDownAction.mouse1 = function()
		if not mouseLook then
			if not showUI or uiRoot:mouseDown( mouseX / ScreenWidth, mouseY / ScreenHeight ) == uiRoot then
				-- Clicked in the window, not on a UI element
				mouseLook = true;
				ignoreNextMM = true;
				mouseX, mouseY = math.floor( ScreenWidth / 2 ), math.floor( ScreenHeight / 2 );
				eihort.warpMouse( mouseX, mouseY );
				eihort.showCursor( false );
				if showUI then
					uiRoot:mouseLeave();
				end
			end
		end
	end;
	EventUpAction.mouse1 = function()
		if not mouseLook and showUI then
			uiRoot:mouseUp( mouseX / ScreenWidth, mouseY / ScreenHeight );
		end
	end;

	-------------------------------------------------------------------
	-- Main UI
	do
		local panelW = 0.28;
		local panelR = 0.98;
		local panelX = panelR - panelW;
		local panelY = 0.1;
		local extraOptionsPanel = uiRoot:addComposite( 1, 0, 1, 1 );
		local fontHt = 0.03;
		local buttonHt = 0.035;
		local buttonSpace = buttonHt + 0.01;
		local y = panelY;
		local inDim = 0;

		-- Show Extra Options toggle
		uiRoot:addToggle(
			function() extraOptionsPanel.invisible = false; end, 
			function() extraOptionsPanel.invisible = true; end,
			"Extra Options", fontHt*1.2, 0, panelX-0.1, 0.02, panelW+0.1, 0.04 );
		extraOptionsPanel.invisible = true;

		-- Move spawn button
		local moveSpawnButton = extraOptionsPanel:addButton(
			function() execWhenNothingIsLoading( function()
				worldView:destroy();
				world:destroy();
				blocks:destroy();
				beginWorldMenu();
			end ); end,
			"Select another world", fontHt, 0, panelX, y, panelW, buttonHt );
		y = y + buttonSpace;

		-- Move spawn button
		local moveSpawnButton = extraOptionsPanel:addButton(
			function() moveSpawnHere( world, inDim, eyeX, eyeY, eyeZ ); end,
			"Move spawn here", fontHt, 0, panelX, y, panelW, buttonHt );
		y = y + buttonSpace;

		-- Move player button
		local movePlayerButton = extraOptionsPanel:addButton(
			function() movePlayerHere( world, inDim, eyeX, eyeY, eyeZ ); end,
			"Move player here", fontHt, 0, panelX, y, panelW, buttonHt );
		y = y + buttonSpace;

		-- Nether/Overworld/End switching
		local function moveToDim( dim )
			if inDim ~= dim then
				if dim == -1 then
					-- To nether
					eyeX = eyeX / 8;
					eyeZ = eyeZ / 8;
				elseif inDim == -1 then
					-- From nether
					eyeX = eyeX * 8;
					eyeZ = eyeZ * 8;
				end
				refreshPosition();
				if dim == 0 then
					world:changeRootPath( worldPath );
				else
					world:changeRootPath( worldPath .. "DIM" .. dim .. "/" );
				end
				worldView:reloadAll();
				inDim = dim;
			end
		end
		local dimButtonW = (panelW - 2*(buttonSpace - buttonHt)) / 3;
		extraOptionsPanel:addButton(
			function() execWhenNothingIsLoading( moveToDim, -1 ); end,
			"Nether", fontHt, 0, panelX, y, dimButtonW, buttonHt );
		extraOptionsPanel:addButton(
			function() execWhenNothingIsLoading( moveToDim, 0 ); end,
			"Overworld", fontHt*0.95, 0, panelX+dimButtonW+buttonSpace-buttonHt, y, dimButtonW, buttonHt );
		extraOptionsPanel:addButton(
			function() execWhenNothingIsLoading( moveToDim, 1 ); end,
			"The End", fontHt, 0, panelX+2*(dimButtonW+buttonSpace-buttonHt), y, dimButtonW, buttonHt );
		y = y + buttonSpace;

		-- No block lighting
		local function setNoBlockLighting( on )
			blocks:noBlockLighting( on );
			worldView:reloadAll();
		end
		local noBlockLighting = extraOptionsPanel:addToggle(
			function() execWhenNothingIsLoading( setNoBlockLighting, true ); end,
			function() execWhenNothingIsLoading( setNoBlockLighting, false ); end,
			"No block lighting", fontHt, 0, panelX, y, panelW, buttonHt );
		noBlockLighting.alt = "Disables light from torches and other light-emitting blocks.\nThis makes block highlights more visible";
		y = y + buttonSpace;

		local hiliteW = (panelW + buttonHt - buttonSpace) / 2;
		local hiliteX2 = panelX + hiliteW + buttonSpace - buttonHt;
		local function addBlockHighlight( i, name, ... )
			local ids = { ... };
			local function changeHighlight( on )
				for _, id in ipairs( ids ) do
					blocks:setHighlight( id, on );
				end
				worldView:reloadAll();
			end
			local x = panelX;
			if math.fmod( i, 2 ) == 1 then
				x = hiliteX2;
			end

			extraOptionsPanel:addToggle(
				function() execWhenNothingIsLoading( changeHighlight, true ); end,
				function() execWhenNothingIsLoading( changeHighlight, false ); end,
				name, fontHt, 0, x, y, hiliteW, buttonHt );

			if math.fmod( i, 2 ) == 0 then
				y = y + buttonSpace;
			end
		end
		for i, v in ipairs( blockHighlights ) do
			addBlockHighlight( i, unpack( v ) );
		end
		if math.fmod( #blockHighlights, 2 ) == 1 then
			y = y + buttonSpace;
		end

		extraOptionsPanel.rect:setRect( panelX-0.01, panelY-0.01, panelW+0.02, y+buttonHt-buttonSpace-panelY+0.02 );
	end

	-------------------------------------------------------------------
	-- Screen stuff
	EventDownAction.mfocus = nil;
	EventUpAction.mfocus = function()
		if mouseLook then
			mouseLook = false;
			eihort.showCursor( true );
			if showUI then
				uiRoot:mouseEnter( mouseX / ScreenWidth, mouseY / ScreenHeight );
			end
		end
	end;
	Event.redraw = function()
		eihort.beginRender();

		sky:render( worldView, 0, (eyeY - 32) / viewDistance, 0 );
		if eyeY > 32 then
			worldView:setFog( viewDistance * 0.8, viewDistance, sky:getHorizFogColor() );
		else
			worldView:setFog( viewDistance * 0.8, viewDistance, 0, 0, 0 );
		end
		worldView:render( true );

		if takeScreenshot then
			takeScreenshot = false;
			takeScreenshotNow();
		end

		if showUI then
			local ctx = eihort.newUIContext();
			uiRoot:render( ctx );
			refreshInfoDisplay();
			infoDisplay:render( ctx );
			ctx:destroy();
		end

		assert( eihort.endRender() );
	end;
	EventDownAction.resize = function()
		resetCameraParams();
		uiRoot:changedAspect();
		infoDisplay:changedAspect();
	end;

	-------------------------------------------------------------------
	-- Movement controls
	local function moveEye( dx, dy, dz, scale )
		eyeX = eyeX + dx * scale;
		eyeY = eyeY + dy * scale;
		eyeZ = eyeZ + dz * scale;
	end;
	local function frameMove( dt )
		if execNoLoad[1] and not worldView:isLoading() then
			worldView:pauseLoading( false );
			local e = execNoLoad;
			execNoLoad = { };
			for _, v in ipairs( e ) do
				v();
			end
		end
		if controlsOn > 0 then
			local speed = dt * speedModifier;
			local mvSpeed = (Config.movement_speed or 50) * speed;
			if controls.moveforward > 0 then
				moveEye( fwdX, fwdY, fwdZ, mvSpeed );
			end
			if controls.moveback > 0 then
				moveEye( fwdX, fwdY, fwdZ, -mvSpeed );
			end
			if controls.moveright > 0 then
				moveEye( rightX, rightY, rightZ, mvSpeed );
			end
			if controls.moveleft > 0 then
				moveEye( rightX, rightY, rightZ, -mvSpeed );
			end
			if controls.moveup > 0 then
				moveEye( upX, upY, upZ, mvSpeed );
			end
			if controls.movedown > 0 then
				moveEye( upX, upY, upZ, -mvSpeed );
			end
			if controls['viewdist+'] > 0 or controls['viewdist-'] > 0 then
				local vdSpeed = speed * (Config.view_distance_speed or 50);
				if controls['viewdist+'] > 0 then
					viewDistance = viewDistance + vdSpeed;
					if viewDistance > 10000 then
						viewDistance = 10000;
					end
				end
				if controls['viewdist-'] > 0 then
					viewDistance = viewDistance - vdSpeed;
					if viewDistance < 50 then
						viewDistance = 50;
					end
				end
				resetCameraParams();
				worldView:setViewDistance( viewDistance );
			end
			if controls['time+'] > 0 or controls['time-'] > 0 then
				local tSpeed = speed * (Config.time_speed or 6) / 12;
				if controls['time+'] > 0 then
					worldTime = worldTime + tSpeed;
					if worldTime > 1 then
						worldTime = worldTime - 2;
					end
				end
				if controls['time-'] > 0 then
					worldTime = worldTime - tSpeed;
					if worldTime < -1 then
						worldTime = worldTime + 2;
					end
				end
				worldView:setOutsideLightLevel( sky:setTime( worldTime ) );
			end
			refreshPosition();
			return true;
		end
		return false;
	end;

	-------------------------------------------------------------------
	-- Idle controller
	local minFrameTime = 1 / (Config.fps_limit or 60);
	local lastFrameTime = eihort.getTime();

	Event.idle = function()
		local t = eihort.getTime();
		local dt = t - lastFrameTime;
		if dt > minFrameTime then
			if dt > 0.1 then
				dt = 0.1;
				lastFrameTime = t;
			else
				lastFrameTime = lastFrameTime + dt;
			end
			return not frameMove( dt );
		end
		return true;
	end;
end

